import pdb
from bagle import model
import numpy as np
import matplotlib.pyplot as plt
from bagle import frame_convert as fc
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.table import Table

from matplotlib import logging
logging.getLogger('matplotlib.font_manager').disabled = True

def blah():
    t = Table.read('/u/jlu/work/microlens/OB110462/a_2022_03_01/model_fits/pspl/ogle_hst_phot_gp/base_b/b0_post_equal_weights.dat', format='ascii')
    t = Table.read('/u/jlu/work/microlens/OB110462/a_2021_12_20/model_fits/ogle_hst_phot/base_a/a0_post_equal_weights.dat', format='ascii')
    ra = '17:51:40.19'
    dec = '-29:53:26.3'

    t0_h = t['col1']
    u0_h = t['col2']
    tE_h = t['col3']
    piEE_h = t['col4']
    piEN_h = t['col5']
    t0par = 55763.327

    output_arr = fc.convert_helio_geo_phot(ra, dec, t0_h, u0_h, 
                                           tE_h, piEE_h, piEN_h,
                                           t0par, in_frame='helio',
                                           murel_in='SL', murel_out='LS',
                                           coord_in='EN', coord_out='tb',
                                           plot=False)

    t0_g, u0_g, tE_g, piEE_g, piEN_g = output_arr

    fig, ax = plt.subplots(2, 5, figsize=(10, 6))
    ax[0,0].hist(t0_h)
    ax[0,1].hist(u0_h)
    ax[0,2].hist(tE_h)
    ax[0,3].hist(piEE_h)
    ax[0,4].hist(piEN_h)

    ax[1,0].hist(t0_g)
    ax[1,1].hist(u0_g)
    ax[1,2].hist(tE_g)
    ax[1,3].hist(piEE_g)
    ax[1,4].hist(piEN_g)
    plt.show()

def test_array_input():
    ra = 259.0
    dec = -29.0
    t0_h = 57000 * np.ones(8)
    u0_h = [0.5, 0.5, 0.5, 0.5, -0.5, -0.5, -0.5, -0.5]
    tE_h = 300 #* np.ones(8)
    piEE_h = np.array([0.2, -0.2, 0.2, -0.2, 0.2, -0.2, 0.2, -0.2])
    piEN_h = np.array([0.1, 0.1, -0.1, -0.1, 0.1, 0.1, -0.1, -0.1])
    t0par = 57100

    output_arr = fc.convert_helio_geo_phot(ra, dec, t0_h, u0_h, 
                                       tE_h, piEE_h, piEN_h,
                                       t0par, in_frame='helio',
                                       murel_in='SL', murel_out='LS',
                                       coord_in='EN', coord_out='tb',
                                       plot=False)

    t0_g_arr, u0_g_arr, tE_g_arr, piEE_g_arr, piEN_g_arr = output_arr

    for ii in np.arange(8):
        print(ii)
        output = fc.convert_helio_geo_phot(ra, dec, t0_h[ii], u0_h[ii], 
                                           tE_h,#[ii], 
                                           piEE_h[ii], piEN_h[ii],
                                           t0par, in_frame='helio',
                                           murel_in='SL', murel_out='LS',
                                           coord_in='EN', coord_out='tb',
                                           plot=False)

        t0_g, u0_g, tE_g, piEE_g, piEN_g = output
        
        assert t0_g == t0_g_arr[ii]
        assert u0_g == u0_g_arr[ii]
        assert tE_g == tE_g_arr[ii]
        assert piEE_g == piEE_g_arr[ii]
        assert piEN_g == piEN_g_arr[ii]
        
def compare_mulens_to_bagle_psbl_phot(ra, dec, t_mjd,
                                      t0_m, u0_m, tE_m,
                                      piEE_m, piEN_m, t0par,
                                      q_m, alpha_m, sep_m,
                                      return_mag=False,
                                      plot_lc=False,
                                      plot_conv=False,
                                      verbose=False):
    """
    return_mag : bool
        True : lightcurve in  magnitudes assuming an unblended 22nd mag source.
        False : lightcurve in magnification units.
    """

    output = fc.convert_bagle_mulens_psbl_phot(ra, dec,
                                               t0_m, u0_m, tE_m,
                                               piEE_m, piEN_m, t0par,
                                               q_m, alpha_m, sep_m,
                                               mod_in='mulens', plot=plot_conv)
    t0_b, u0_b, tE_b, piEE_b, piEN_b, q_b, alpha_b, sep_b = output

    # Get HJD from MJD (since MulensModel uses HJD).
    t_hjd = t_mjd + 2400000.5

    # Turn RA and Dec into SkyCoord object (for MulensModel).
    coords = SkyCoord(ra, dec, unit=(u.deg, u.deg))

    # Get the lightcurve from the geoprojected parameters.
    mag_mulens = get_phot_mulens_psbl(coords, t0_m, u0_m, tE_m,
                                      piEE_m, piEN_m, t0par,
                                      alpha_m, q_m, sep_m, t_hjd,
                                      return_mag=return_mag)

    mag_pylima = get_phot_pylima_psbl(ra, dec, t0_m, u0_m, tE_m,
                                      piEE_m, piEN_m, t0par,
                                      alpha_m, q_m, sep_m, t_hjd,
                                      return_mag=return_mag)

    # Get the lightcurve from the heliocentric parameters.
    mag_bagle, mod_bagle = get_phot_bagle_psbl(ra, dec, t0_b, u0_b,
                                               tE_b, piEE_b, piEN_b,
                                               alpha_b, q_b, sep_b, t_mjd,
                                               return_mag=return_mag)

    if verbose:
        print('Mulens Params:')
        print(t0_m, u0_m, tE_m,
              piEE_m, piEN_m, t0par,
              q_m, alpha_m, sep_m)

        print('PyLIMA Params:')
        print(t0_m, u0_m, tE_m,
              piEE_m, piEN_m, t0par,
              q_m, alpha_m, sep_m)

        print('BAGLE Params:')
        print(t0_b, u0_b, tE_b,
              piEE_b, piEN_b,
              q_b, alpha_b, sep_b)

    diff_mulens = mag_mulens - mag_bagle
    diff_pylima = mag_pylima - mag_bagle

    if plot_lc:
        fig, ax = plt.subplots(3, 1, figsize=(8,8), sharex=True, num=3)
        plt.clf()
        fig, ax = plt.subplots(3, 1, figsize=(8,8), sharex=True, num=3)
        ax[0].plot(t_mjd, mag_mulens, '-', label='M')
        ax[0].plot(t_mjd, mag_pylima, '-', label='P')
        ax[0].plot(t_mjd, mag_bagle, '-', label='B')
        ax[0].legend()
        ax[1].plot(t_mjd, diff_mulens, '.')
        ax[2].plot(t_mjd, diff_pylima, '.')
        if return_mag:
            ax[0].set_ylabel('Mag')
            ax[1].set_ylabel('Mag Mulens - BAGLE', fontsize=12)
            ax[2].set_ylabel('Mag pyLIMA - BAGLE', fontsize=12)
            ax[0].invert_yaxis()
        else:
            ax[0].set_ylabel('Amp')
            ax[1].set_ylabel('Amp Mulens - BAGLE', fontsize=12)
            ax[2].set_ylabel('Amp pyLIMA - BAGLE', fontsize=12)
            ax[0].set_ylim(0.95, 10)
            ax[1].set_ylim(-0.4, 0.4)
            ax[2].set_ylim(-0.4, 0.4)
        ax[2].set_xlabel('MJD')

        # Print mean and max difference
        ax[1].set_title(f'Diff mean={diff_mulens.mean():.3e}, max={np.abs(diff_mulens).max():.3e}',
                        fontsize=12)
        ax[2].set_title(f'Diff mean={diff_pylima.mean():.3e}, max={np.abs(diff_pylima).max():.3e}',
                        fontsize=12)


    # Make sure that the conversion works by asserting
    # that the lightcurves are no more different than
    # 1e-1 magnitudes on average.
    diff = (mag_mulens - mag_bagle)/mag_mulens
    total_diff = np.sum(np.abs(diff))
    assert total_diff/len(t_mjd) < 1e-4

    return


def compare_bagle_to_mulens_psbl_phot(ra, dec, t_mjd,
                                      t0_b, u0_b, tE_b,
                                      piEE_b, piEN_b, t0par,
                                      q_b, alpha_b, sep_b,
                                      return_mag=False,
                                      plot_lc=False,
                                      plot_conv=False,
                                      verbose=False, fignum=3):
    
    # Get the BAGLE lightcurve from the heliocentric parameters.
    mag_bagle, mod_bagle = get_phot_bagle_psbl(ra, dec, t0_b, u0_b,
                                               tE_b, piEE_b, piEN_b,
                                               alpha_b, q_b, sep_b, t_mjd,
                                               return_mag=return_mag)

    output = fc.convert_bagle_mulens_psbl_phot(ra, dec,
                                               t0_b, u0_b, tE_b,
                                               piEE_b, piEN_b, t0par,
                                               q_b, alpha_b, sep_b,
                                               mod_in='bagle', plot=plot_conv)
    t0_m, u0_m, tE_m, piEE_m, piEN_m, q_m, alpha_m, sep_m = output

    output = fc.convert_bagle_pylima_psbl_phot(ra, dec,
                                               t0_b, u0_b, tE_b,
                                               piEE_b, piEN_b, t0par,
                                               q_b, alpha_b, sep_b,
                                               mod_in='bagle', plot=plot_conv)
    t0_p, u0_p, tE_p, piEE_p, piEN_p, q_p, alpha_p, sep_p = output

    # Get HJD from MJD (since MulensModel uses HJD).
    t_hjd = t_mjd + 2400000.5

    # Turn RA and Dec into SkyCoord object (for MulensModel).
    coords = SkyCoord(ra, dec, unit=(u.deg, u.deg))

    # Get the lightcurve from the geoprojected parameters.
    mag_mulens = get_phot_mulens_psbl(coords, t0_m, u0_m, tE_m,
                                      piEE_m, piEN_m, t0par,
                                      alpha_m, q_m, sep_m, t_hjd,
                                      return_mag=return_mag)

    mag_pylima = get_phot_pylima_psbl(ra, dec, t0_p, u0_p, tE_p,
                                      piEE_p, piEN_p, t0par,
                                      alpha_p, q_p, sep_p, t_hjd,
                                      return_mag=return_mag)

    if verbose:
        print('Mulens Params:')
        print(t0_m, u0_m, tE_m,
              piEE_m, piEN_m, t0par,
              q_m, alpha_m, sep_m)

        print('PyLIMA Params:')
        print(t0_p, u0_p, tE_p,
              piEE_p, piEN_p, t0par,
              q_p, alpha_p, sep_p)

        print('BAGLE Params:')
        print(t0_b, u0_b, tE_b,
              piEE_b, piEN_b,
              q_b, alpha_b, sep_b)

    diff_mulens = mag_mulens - mag_bagle
    diff_pylima = mag_pylima - mag_bagle

    if plot_lc:
        fig, ax = plt.subplots(3, 1, figsize=(8,8), sharex=True, num=fignum)
        plt.clf()
        fig, ax = plt.subplots(3, 1, figsize=(8,8), sharex=True, num=fignum)
        ax[0].plot(t_mjd, mag_mulens, '-', label='M')
        ax[0].plot(t_mjd, mag_pylima, '-', label='P')
        ax[0].plot(t_mjd, mag_bagle, '-', label='B')
        ax[0].legend()
        ax[1].plot(t_mjd, diff_mulens, '-')
        ax[2].plot(t_mjd, diff_pylima, '-')
        if return_mag:
            ax[0].set_ylabel('Mag')
            ax[1].set_ylabel('Mag Mulens - Bagle', fontsize=12)
            ax[2].set_ylabel('Mag pyLIMA - BAGLE', fontsize=12)
            ax[0].invert_yaxis()
        else:
            ax[0].set_ylabel('Amp')
            ax[1].set_ylabel('Amp Mulens - BAGLE', fontsize=12)
            ax[2].set_ylabel('Amp pyLIMA - BAGLE', fontsize=12)
            ax[0].set_ylim(0.95, 10)
            ax[1].set_ylim(-0.4, 0.4)
            ax[2].set_ylim(-0.4, 0.4)
        ax[2].set_xlabel('MJD')

        # Print mean and max difference
        ax[1].set_title(f'Diff mean={diff_mulens.mean():.3e}, max={np.abs(diff_mulens).max():.3e}',
                        fontsize=12)
        ax[2].set_title(f'Diff mean={diff_pylima.mean():.3e}, max={np.abs(diff_pylima).max():.3e}',
                        fontsize=12)



#        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
#        plt.clf()
#        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
#        ax[0].plot(t_mjd, mag_pylima, '.', ms=10)
#        ax[0].plot(t_mjd, mag_bagle, '.', ms=5)
#        ax[0].plot(t_mjd, mag_mulens, '.', ms=3)
#        ax[0].invert_yaxis()
#        ax[1].plot(t_mjd, mag_mulens - mag_pylima, '.', ms=10, label='M-P')
#        ax[1].plot(t_mjd, mag_mulens - mag_bagle, '.', ms=5, label='M-B')
#        ax[0].set_ylabel('Mag')
#        ax[1].set_ylabel('Mag Diff')
#        ax[1].set_xlabel('MJD')
#        ax[1].legend()
#        plt.show()
#        plt.pause(0.5)

    # Make sure that the conversion works by asserting
    # that the lightcurves are no more different than
    # 1e-1 magnitudes on average.
    diff = (diff_mulens) / mag_mulens
    total_diff = np.sum(np.abs(diff))
    assert total_diff/len(t_mjd) < 1e-4


def compare_bagle_geo_to_helio_phot(ra, dec, t_mjd,
                                    t0_g, u0_g, tE_g,
                                    piEE_g, piEN_g,
                                    t0par,
                                    plot_conv=False,
                                    plot_lc=False,
                                    verbose=False):

    """
    Test conversion from geocentric projected frame
    (using lens-source tau-beta convention) to
    heliocentric frame (using source-lens East-North 
    convention) photometry parameters in BAGLE.
    
    Parameters
    ----------
    ra, dec : int or float (degrees)

    t_mjd, t0_g, t0par need to be in MJD.

    t0_g, u0_g, tE_g, piEE_g, piEN_g need to be
    in geocentric projected frame using lens-source
    and tau-beta conventions.
    """
    # Convert the geoprojected parameters (in lens-source tau-beta)
    # to heliocentric parameters (in source-lens East-North).
    output = fc.convert_helio_geo_phot(ra, dec, t0_g, u0_g, 
                                       tE_g, piEE_g, piEN_g,
                                       t0par, in_frame='geo',
                                       murel_in='LS', murel_out='SL',
                                       coord_in='tb', coord_out='EN',
                                       plot=plot_conv)

    t0_h, u0_h, tE_h, piEE_h, piEN_h = output

    # Get the lightcurve from the geoprojected parameters.
    mag_bagle_geoproj = get_phot_bagle_geoproj(ra, dec, t0_g, u0_g, 
                                               tE_g, piEE_g, piEN_g, 
                                               t_mjd, t0par)

    # Get the lightcurve from the heliocentric parameters.
    mag_bagle = get_phot_bagle(ra, dec, t0_h, u0_h, 
                               tE_h, piEE_h, piEN_h, t_mjd)

    if verbose:
        label = ['t0 (MJD)', 'u0', 'tE (days)', 'piEE', 'piEN', 't0par']
        geoproj = [t0_g, u0_g, tE_g, piEE_g, piEN_g, t0par]
        helio = [t0_h, u0_h, tE_h, piEE_h, piEN_h]

        print('*******************************************')
        print('Geo proj parameters, tau-beta, L-S frame')
        for gg, _ in enumerate(geoproj):
            print('{0} : {1:.2f}'.format(label[gg], geoproj[gg]))
        print('-------------------------------------------')
        print('Helio parameters, East-North, S-L frame')
        for hh, _ in enumerate(helio):
            print('{0} : {1:.2f}'.format(label[hh], helio[hh]))
        print('*******************************************')

    if plot_lc:
        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
        plt.clf()
        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
        ax[0].plot(t_mjd, mag_bagle_geoproj, 'o', label='Geo proj')
        ax[0].plot(t_mjd, mag_bagle, '.', label='Helio')
        ax[0].legend()
        ax[0].invert_yaxis()
        ax[1].plot(t_mjd, mag_bagle_geoproj - mag_bagle, '.')
        ax[0].set_ylabel('Mag')
        ax[1].ticklabel_format(axis='y', style='sci', scilimits=(0,0))
        ax[1].set_ylabel('Geo proj - Helio')
        ax[1].set_xlabel('MJD')
        plt.show()
        plt.pause(0.5)

    # Make sure that the conversion works by asserting
    # that the lightcurves are no more different than
    # 1e-4 magnitudes on average.
    # Note: current test fails if require < 1e-5.
    diff = (mag_bagle_geoproj - mag_bagle)/mag_bagle_geoproj
    total_diff = np.sum(np.abs(diff))
    assert total_diff/len(t_mjd) < 1e-4

def compare_bagle_helio_to_geo_phot(ra, dec, t_mjd,
                                    t0_h, u0_h, tE_h,
                                    piEE_h, piEN_h,
                                    t0par,
                                    plot_conv=False,
                                    plot_lc=False,
                                    verbose=False,
                                    fignum=3):
    """
    Test conversion from heliocentric frame (using 
    source-lens East-North convention) to geocentric 
    projected frame (using lens-source tau-beta convention)
    photometry parameters in BAGLE.

    Parameters
    ----------
    ra, dec : int or float (degrees)

    t_mjd, t0_h, t0par need to be in MJD.
    
    t0_h, u0_h, tE_h, piEE_h, piEN_h need to be
    in heliocentric frame using source-lens
    and East-North conventions.
    """
    # Convert the heliocentric parameters (in source-lens East-North)
    # to geoprojected parameters (in lens-source tau-beta).
    output = fc.convert_helio_geo_phot(ra, dec, t0_h, u0_h, 
                                       tE_h, piEE_h, piEN_h,
                                       t0par, in_frame='helio',
                                       murel_in='SL', murel_out='LS',
                                       coord_in='EN', coord_out='tb',
                                       plot=plot_conv)

    t0_g, u0_g, tE_g, piEE_g, piEN_g = output

    # Get the lightcurve from the geoprojected parameters.
    mag_bagle_geoproj = get_phot_bagle_geoproj(ra, dec, t0_g, u0_g, 
                                               tE_g, piEE_g, piEN_g, 
                                               t_mjd, t0par)

    # Get the lightcurve from the heliocentric parameters.
    mag_bagle = get_phot_bagle(ra, dec, t0_h, u0_h, 
                               tE_h, piEE_h, piEN_h, t_mjd)

    if verbose:
        label = ['t0 (MJD)', 'u0', 'tE (days)', 'piEE', 'piEN', 't0par']
        geoproj = [t0_g, u0_g, tE_g, piEE_g, piEN_g, t0par]
        helio = [t0_h, u0_h, tE_h, piEE_h, piEN_h]

        print('*******************************************')
        print('Geo proj parameters, tau-beta, L-S frame')
        for gg, _ in enumerate(geoproj):
            print('{0} : {1:.2f}'.format(label[gg], geoproj[gg]))
        print('-------------------------------------------')
        print('Helio parameters, East-North, S-L frame')
        for hh, _ in enumerate(helio):
            print('{0} : {1:.2f}'.format(label[hh], helio[hh]))
        print('*******************************************')

    if plot_lc:
        fig, ax = plt.subplots(2, 1, sharex=True, num=fignum)
        plt.clf()
        fig, ax = plt.subplots(2, 1, sharex=True, num=fignum)
        ax[0].plot(t_mjd, mag_bagle_geoproj, 'o', label='Geo proj')
        ax[0].plot(t_mjd, mag_bagle, '.', label='Helio')
        ax[0].invert_yaxis()
        ax[0].legend()
        ax[1].plot(t_mjd, mag_bagle_geoproj - mag_bagle, '.')
        ax[0].set_ylabel('Mag')
        ax[1].ticklabel_format(axis='y', style='sci', scilimits=(0,0))
        ax[1].set_ylabel('Geo proj - Helio')
        ax[1].set_xlabel('MJD')
        plt.show()
        plt.pause(1)

    # Make sure that the conversion works by asserting
    # that the lightcurves are no more different than
    # 1e-5 magnitudes on average.
    # Note: current test fails if require < 1e-6.
    diff = (mag_bagle - mag_bagle_geoproj)/mag_bagle
    total_diff = np.sum(np.abs(diff))
    assert total_diff/len(t_mjd) < 1e-5
    
def test_convert_u0_t0_psbl():
    """
    Test converting u0 and t0 from primary center coordinate system
    to geometric center coordinate frame. This is done via plugging parameters
    into PSBL_PhotAstrom_Par_Param7, converting them mathematically, then
    plugging them into PSBL_PhotAstrom_Par_Param1, then checking photometry
    is equal.
    """
    mLp = 10
    mLs = 10
    xS0_E = 0 
    xS0_N = 0 
    muL_E = 0
    muL_N = 0
    muS_E = 0
    muS_N = 2
    dL = 4000
    dS = 8000
    sep = 0.1
    alpha = 40
    b_sff = 0.5
    mag_src = 18
    raL = 270
    decL = -30
    dmag_Lp_Ls = 20.0
    
    t0_p = 0
    beta_p = 0.5
    psbl_model = model.PSBL_PhotAstrom_Par_Param7(mLp, mLs, t0_p, xS0_E, xS0_N, beta_p, 
                                                  muL_E, muL_N, muS_E, muS_N, dL, dS,
                                                  sep, alpha, [b_sff], [mag_src], [dmag_Lp_Ls],
                                                  raL=raL, decL=decL, 
                                                  root_tol = 0.00000001)
    
    # Conversions determined by specific geometry (can be checked in u0/t0 conversion overleaf)
    # https://www.overleaf.com/read/qzvfpvkpxdqm
    t0_geom = psbl_model.t0_p - psbl_model.tE*np.cos(np.deg2rad(psbl_model.phi))*(psbl_model.sep/(2*psbl_model.thetaE_amp))
    beta_geom = (psbl_model.u0_amp_p - (-1)*(psbl_model.sep/(2*psbl_model.thetaE_amp))*np.sin(np.deg2rad(psbl_model.phi)))*psbl_model.thetaE_amp
    
    psbl_model_geom = model.PSBL_PhotAstrom_Par_Param1(mLp, mLs, t0_geom, xS0_E, xS0_N, beta_geom, 
                                                       muL_E, muL_N, muS_E, muS_N, dL, dS,
                                                       sep, alpha, [b_sff], [mag_src], [dmag_Lp_Ls],
                                                       raL=raL, decL=decL, 
                                                       root_tol = 0.00000001)
    
    
    dt = np.linspace(-500, 500, 200)
    img, amp = psbl_model.get_all_arrays(dt)
    phot = psbl_model.get_photometry(dt, amp_arr=amp)
    
    img_geom, amp_geom = psbl_model_geom.get_all_arrays(dt)
    phot_geom = psbl_model_geom.get_photometry(dt, amp_arr=amp_geom)
    
    assert(np.allclose(phot, phot_geom, rtol=1e-15))


def compare_bagle_to_mulens(ra, dec, t_mjd,
                            t0_h, u0_h, tE_h,
                            piEE_h, piEN_h,
                            t0par,
                            plot_conv=False,
                            plot_lc=False,
                            verbose=False):
    """
    Test conversion from heliocentric frame (using 
    source-lens East-North convention) in BAGLE to geocentric 
    projected frame (using lens-source tau-beta convention)
    in MulensModel for photometry parameters.

    Parameters
    ----------
    ra, dec : int or float (degrees)

    t_mjd, t0_h, t0par need to be in MJD.

    t0_h, u0_h, tE_h, piEE_h, piEN_h need to be
    in heliocentric frame using source-lens
    and East-North conventions.
    """
    # Convert the heliocentric parameters (in source-lens East-North)
    # to geoprojected parameters (in lens-source tau-beta).
    output = fc.convert_helio_geo_phot(ra, dec, t0_h, u0_h, 
                                       tE_h, piEE_h, piEN_h,
                                       t0par, in_frame='helio',
                                       murel_in='SL', murel_out='LS',
                                       coord_in='EN', coord_out='tb',
                                       plot=plot_conv)

    t0_g, u0_g, tE_g, piEE_g, piEN_g = output

    # Get HJD from MJD (since MulensModel uses HJD).
    t_hjd = t_mjd + 2400000.5
    
    # Turn RA and Dec into SkyCoord object (for MulensModel).
    coords = SkyCoord(ra, dec, unit=(u.deg, u.deg))

    # Get the lightcurve from the geoprojected parameters.
    mag_mulens = get_phot_mulens(coords, t0_g, u0_g, 
                                 tE_g, piEE_g, piEN_g, 
                                 t0par, t_hjd)

    # Get the lightcurve from the heliocentric parameters.
    mag_bagle = get_phot_bagle(ra, dec, t0_h, u0_h, 
                               tE_h, piEE_h, piEN_h, t_mjd)
    
    if verbose:
        label = ['t0 (MJD)', 'u0', 'tE (days)', 'piEE', 'piEN', 't0par']
        geoproj = [t0_g, u0_g, tE_g, piEE_g, piEN_g, t0par]
        helio = [t0_h, u0_h, tE_h, piEE_h, piEN_h]

        print('*******************************************')
        print('Geo proj parameters, tau-beta, L-S frame')
        for gg, _ in enumerate(geoproj):
            print('{0} : {1:.2f}'.format(label[gg], geoproj[gg]))
        print('-------------------------------------------')
        print('Helio parameters, East-North, S-L frame')
        for hh, _ in enumerate(helio):
            print('{0} : {1:.2f}'.format(label[hh], helio[hh]))
        print('*******************************************')

    if plot_lc:
        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
        plt.clf()
        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
        ax[0].plot(t_mjd, mag_mulens, 'o', label='Mulens')
        ax[0].plot(t_mjd, mag_bagle, '.', label='BAGLE')
        ax[0].invert_yaxis()
        ax[0].legend()
        ax[1].plot(t_mjd, mag_mulens - mag_bagle, '.')
        ax[0].set_ylabel('Mag')
        ax[1].ticklabel_format(axis='y', style='sci', scilimits=(0,0))
        ax[1].set_ylabel('Mulens - BAGLE')
        ax[1].set_xlabel('MJD')
        plt.show()
        plt.pause(0.5)

    # Make sure that the conversion works by asserting
    # that the lightcurves are no more different than
    # 1e-5 magnitudes on average.
    # Note: current test fails if require < 1e-6.
    diff = (mag_bagle - mag_mulens)/mag_bagle
    total_diff = np.sum(np.abs(diff))
    assert total_diff/len(t_mjd) < 1e-5

def compare_mulens_to_bagle(ra, dec, t_mjd,
                            t0_g, u0_g, tE_g,
                            piEE_g, piEN_g,
                            t0par,
                            plot_conv=False,
                            plot_lc=False,
                            verbose=False):
    """
    Test conversion from geocentric projected frame (using 
    lens-source tau-beta convention) in MulensModel to 
    heliocentric frame (using source-lens East-North 
    convention) in BAGLE for photometry parameters.

    Parameters
    ----------
    ra, dec : int or float (degrees)

    t_mjd, t0_g, t0par need to be in MJD.

    t0_g, u0_g, tE_g, piEE_g, piEN_g need to be
    in geocentric projected frame using lens-source
    and tau-beta conventions.
    """
    # Convert the geoprojected parameters (in lens-source tau-beta)
    # to heliocentric parameters (in source-lens East-North).
    output = fc.convert_helio_geo_phot(ra, dec, t0_g, u0_g, 
                                       tE_g, piEE_g, piEN_g,
                                       t0par, in_frame='geo',
                                       murel_in='LS', murel_out='SL',
                                       coord_in='tb', coord_out='EN',
                                       plot=plot_conv)

    t0_h, u0_h, tE_h, piEE_h, piEN_h = output

    # Get HJD from MJD (since MulensModel uses HJD).
    t_hjd = t_mjd + 2400000.5

    # Turn RA and Dec into SkyCoord object (for MulensModel).
    coords = SkyCoord(ra, dec, unit=(u.deg, u.deg))

    # Get the lightcurve from the heliocentric parameters.
    mag_mulens = get_phot_mulens(coords, t0_g, u0_g, 
                                 tE_g, piEE_g, piEN_g, 
                                 t0par, t_hjd)

    # Get the lightcurve from the geoprojected parameters.
    mag_bagle = get_phot_bagle(ra, dec, t0_h, u0_h, 
                               tE_h, piEE_h, piEN_h, t_mjd)

    if verbose:
        label = ['t0 (MJD)', 'u0', 'tE (days)', 'piEE', 'piEN', 't0par']
        geoproj = [t0_g, u0_g, tE_g, piEE_g, piEN_g, t0par]
        helio = [t0_h, u0_h, tE_h, piEE_h, piEN_h]

        print('*******************************************')
        print('Geo proj parameters, tau-beta, L-S frame')
        for gg, _ in enumerate(geoproj):
            print('{0} : {1:.2f}'.format(label[gg], geoproj[gg]))
        print('-------------------------------------------')
        print('Helio parameters, East-North, S-L frame')
        for hh, _ in enumerate(helio):
            print('{0} : {1:.2f}'.format(label[hh], helio[hh]))
        print('*******************************************')

    if plot_lc:
        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
        plt.clf()
        fig, ax = plt.subplots(2, 1, sharex=True, num=3)
        ax[0].plot(t_mjd, mag_mulens, 'o', label='Mulens')
        ax[0].plot(t_mjd, mag_bagle, '.', label='BAGLE')
        ax[0].invert_yaxis()
        ax[0].legend()
        ax[1].plot(t_mjd, mag_mulens - mag_bagle, '.')
        ax[0].set_ylabel('Mag')
        ax[1].ticklabel_format(axis='y', style='sci', scilimits=(0,0))
        ax[1].set_ylabel('Mulens - BAGLE')
        ax[1].set_xlabel('MJD')
        plt.show()
        plt.pause(0.5)

    # Make sure that the conversion works by asserting
    # that the lightcurves are no more different than
    # 1e-4 magnitudes on average.
    # Note: current test fails if require < 1e-5.
    diff = (mag_mulens - mag_bagle)/mag_mulens
    total_diff = np.sum(np.abs(diff))
    assert total_diff/len(t_mjd) < 1e-4


def get_phot_pylima_psbl(ra, dec, t0_g, u0_g, tE_g,
                         piEE_g, piEN_g, t0par, 
                         alpha_g, q, sep, t_hjd,
                         return_mag=True):
    """
    Compare pyLIMA to our models with some plots. 

    Input parameters are in our conventions and heliocentric coordinate system.
    """
    from pyLIMA import event
    from pyLIMA import telescopes
    from pyLIMA import toolbox as microltoolbox
    from pyLIMA.models import generate_model

    #####
    # PyLIMA setup stuff.
    #####
    # Make fake data for pyLIMA... need this for time array definition.
    pylima_data = np.zeros((len(t_hjd), 3))
    pylima_data[:,0] = t_hjd
    pylima_data[:,1] = np.ones(len(t_hjd))
    pylima_data[:,2] = np.ones(len(t_hjd))

    # Make a telescope object
    pylima_tel = telescopes.Telescope(name='OGLE', camera_filter='I',
                                      lightcurve=pylima_data,
                                      lightcurve_names = ['time', 'mag', 'err_mag'],
                                      lightcurve_units = ['JD', 'mag', 'mag'],
                                      location='Earth',
                                      altitude=1000, longitude=-109.285399, latitude=-27.130)
    pylima_ev = event.Event(ra=ra, dec=dec)
    pylima_ev.name = 'Fubar'
    pylima_ev.telescopes.append(pylima_tel)

    # Set up parameters.
    pylima_t0 = t0_g + 2400000.5
    pylima_u0 = u0_g
    pylima_tE = tE_g
    pylima_piEE = piEE_g 
    pylima_piEN = piEN_g
    pylima_t0_par = t0par + 2400000.5
    pylima_phi = np.deg2rad(alpha_g) # phi in radians
    pylima_q = q
    pylima_s = sep

    # Now finally... create the model.
    pylima_mod = generate_model.create_model('PSBL', pylima_ev,
                                             parallax=['Full', pylima_t0_par],
                                             fancy_parameters=None)

    tmp_params = [pylima_t0, pylima_u0, pylima_tE, 
                  pylima_s, pylima_q, pylima_phi,
                  pylima_piEN, pylima_piEE]
       
    pylima_mod.define_model_parameters()
    pylima_mod.blend_flux_ratio = False
    pylima_mod.astrometry = False
    pylima_mod.orbital_motion_model = ['None']

    mag_src=22
    b_sff=1
    pylima_par = pylima_mod.compute_pyLIMA_parameters(tmp_params)
    pylima_par.fs_OGLE = microltoolbox.brightness_transformation.magnitude_to_flux(mag_src)
    pylima_par.fb_OGLE = pylima_par.fs_OGLE * (1.0 - b_sff) / b_sff

    if return_mag:    
        pylima_lcurve, sf, bf = pylima_mod.compute_the_microlensing_model(pylima_tel, pylima_par)
        pylima_lcurve_mag = microltoolbox.brightness_transformation.flux_to_magnitude(pylima_lcurve)
    else:
        pylima_amp = pylima_mod.model_magnification(pylima_tel, pylima_par)
        pylima_lcurve_mag = pylima_amp

    return pylima_lcurve_mag

def get_phot_mulens(coords, t0_g, u0_g, tE_g, 
                    piEE_g, piEN_g, t0par, t_hjd,
                    return_mag=True):
    """
    Get lightcurve from MulensModel, assuming an 
    unblended 22nd mag source. MulensModel uses the 
    geocentric projected frame, with Gould (tau-beta) 
    and lens-source conventions.
    """
    import MulensModel as mm

    # Set up parameter dictionary. 
    # t_0 and t_0_par need to be in HJD.
    params = {}
    params['t_0'] = t0_g + 2400000.5
    params['t_0_par'] = t0par + 2400000.5
    params['u_0'] = u0_g
    params['t_E'] = tE_g
    params['pi_E_N'] = piEN_g
    params['pi_E_E'] = piEE_g

    # Then instantiate model and get lightcurve.
    my_model = mm.Model(params, coords=coords)

    if return_mag:
        mag_obs = my_model.get_lc(times=t_hjd, source_flux=1, blend_flux=0)
    else:
        mag_obs = my_model.get_magnification(t_hjd)

    return mag_obs


def get_phot_mulens_psbl(coords, t0_g, u0_g, tE_g, 
                         piEE_g, piEN_g, t0par, 
                         alpha_g, q, sep, t_hjd,
                         return_mag=True):
    """
    Get PSBL lightcurve from MulensModel, assuming an 
    unblended 22nd mag source. MulensModel uses the 
    geocentric projected frame, with Gould (tau-beta) 
    and lens-source conventions.
    """
    import MulensModel as mm

    # Set up parameter dictionary. 
    # t_0 and t_0_par need to be in HJD.
    params = {}
    params['t_0'] = t0_g + 2400000.5
    params['t_0_par'] = t0par + 2400000.5
    params['u_0'] = u0_g
    params['t_E'] = tE_g
    params['pi_E_N'] = piEN_g
    params['pi_E_E'] = piEE_g
    params['q'] = q
    params['s'] = sep
    params['alpha'] = alpha_g

    # Then instantiate model and get lightcurve.
    my_model = mm.Model(params, coords=coords)

    if return_mag:
        mag_obs = my_model.get_lc(times=t_hjd, source_flux=1, blend_flux=0)
    else:
        mag_obs = my_model.get_magnification(t_hjd)
    return mag_obs


def get_phot_bagle(ra, dec, t0_h, u0_h, 
                   tE_h, piEE_h, piEN_h, t_mjd,
                   return_mag=True):
    """
    Get lightcurve from BAGLE, assuming an unblended 
    22nd mag source. This BAGLE model uses the 
    heliocentric frame, with Lu (East-North) and
    source-lens conventions.
    """
    # Set up parameter dictionary.
    # t0 needs to be in MJD.
    params = {}
    params['raL'] = ra
    params['decL'] = dec
    params['t0'] = t0_h
    params['u0_amp'] = u0_h
    params['tE'] = tE_h
    params['piE_E'] = piEE_h
    params['piE_N'] = piEN_h
    params['b_sff'] = 1
    params['mag_src'] = 22
    
    # Then instantiate model and get lightcurve.
    mod = model.PSPL_Phot_Par_Param1(params['t0'], params['u0_amp'], 
                                     params['tE'],
                                     params['piE_E'], params['piE_N'],
                                     params['b_sff'], params['mag_src'],
                                     raL=params['raL'], decL=params['decL'])

    if return_mag:
        mag_obs = mod.get_photometry(t_mjd)
    else:
        mag_obs = mod.get_amplification(t_mjd)

    return mag_obs


def get_phot_bagle_psbl(ra, dec, t0_h, u0_h,
                        tE_h, piEE_h, piEN_h, 
                        alpha_h, q, sep, t_mjd,
                        return_mag=True):
    """
    Get PSBL lightcurve from BAGLE, assuming an unblended 
    22nd mag source. This BAGLE model uses the 
    heliocentric frame, with Lu (East-North) and
    source-lens conventions.
    """
    # Set up parameter dictionary.
    # t0 needs to be in MJD.
    params = {}
    params['raL'] = ra
    params['decL'] = dec
    params['t0'] = t0_h
    params['u0_amp'] = u0_h
    params['tE'] = tE_h
    params['piE_E'] = piEE_h
    params['piE_N'] = piEN_h
    params['q'] = q
    params['alpha'] = alpha_h
    params['sep'] = sep
    params['b_sff'] = 1
    params['mag_src'] = 22
    
    # Then instantiate model and get lightcurve.
    # FIXME notational consistency: alpha or phi? (or are they the same?)
    mod = model.PSBL_Phot_Par_Param1(params['t0'], params['u0_amp'], 
                                     params['tE'],
                                     params['piE_E'], params['piE_N'],
                                     params['q'], params['sep'], 
                                     params['alpha'],
                                     params['b_sff'], params['mag_src'],
                                     raL=params['raL'], decL=params['decL'])

    if return_mag:
        mag_obs = mod.get_photometry(t_mjd)
    else:
        mag_obs = mod.get_amplification(t_mjd)

    return mag_obs, mod


def get_phot_bagle_geoproj(ra, dec, t0_h, u0_h, tE_h, 
                           piEE_h, piEN_h, t_mjd, t0par,
                           return_mag=True):
    """
    Get lightcurve from BAGLE, assuming an unblended 
    22nd mag source. This BAGLE model uses the    
    geocentric projected frame, with Gould (tau-beta) 
    and lens-source conventions.
    """
    # Set up parameter dictionary.
    # t0 and t0par need to be in MJD.
    params = {}
    params['raL'] = ra
    params['decL'] = dec
    params['t0'] = t0_h
    params['u0_amp'] = u0_h
    params['tE'] = tE_h
    params['piE_E'] = piEE_h
    params['piE_N'] = piEN_h
    params['b_sff'] = 1
    params['mag_src'] = 22
    params['t0par'] = t0par

    # Then instantiate model and get lightcurve.
    mod = model.PSPL_Phot_Par_Param1_geoproj(params['t0'], params['u0_amp'],
                                             params['tE'],
                                             params['piE_E'], params['piE_N'],
                                             params['b_sff'], params['mag_src'],
                                             params['t0par'],
                                             raL=params['raL'], decL=params['decL'])
    if return_mag:
        mag_obs = mod.get_photometry(t_mjd)
    else:
        mag_obs = mod.get_amplification(t_mjd)

    return mag_obs


def test_bagle_mulens_psbl_phot_set(plot_lc=False, plot_conv=False, verbose=False):

    kwargs = {'plot_lc': plot_lc, 'plot_conv': plot_conv, 'verbose': verbose}

    t_mjd = np.arange(60500 - 500, 60500 + 500, 0.5)

    print('set 1')
    compare_bagle_to_mulens_psbl_phot(259.0, -29.0, t_mjd, 60478, 0.5, 100, 0.2, 0.1, 60500, 0.5, 125, 1.5, **kwargs)
    compare_bagle_to_mulens_psbl_phot(259.0, -29.0, t_mjd, 60478, 0.5, 300, 0.2, 0.1, 60478, 0.5, 90, 1.5, **kwargs)
    compare_bagle_to_mulens_psbl_phot(259.0, -29.0, t_mjd, 60478, 0.3, 50, 0.2, 0.1, 60500, 2, 339, 1.5, **kwargs)
    compare_bagle_to_mulens_psbl_phot(259.0, -29.0, t_mjd, 60478, 0.1, 300, 0.2, 0.1, 60478, 0.5, 90, 1.5, **kwargs)
    compare_bagle_to_mulens_psbl_phot(259.0, -29.0, t_mjd, 60478, 0.1, 50, 0.2, 0.1, 60500, 2, 339, 1.5, **kwargs)

    print('set 2')
    compare_mulens_to_bagle_psbl_phot(259, -29, t_mjd,
                                      60461.11916524347, 0.2580016909695294, 77.22238744939838,
                                      -0.21245185338062486, -0.06974388858629471,
                                60500, 0.5, -116.60896987280785, 1.5, **kwargs)
    compare_mulens_to_bagle_psbl_phot(259.0, -29.0, t_mjd, 60478, 0.5, 300, 0.2, 0.1, 60200, 0.5, 90, 1.5, **kwargs)
    compare_mulens_to_bagle_psbl_phot(259.0, -29.0, t_mjd, 60478, 0.3, 50, 0.2, 0.1, 60300, 2, 339, 1.5, **kwargs)

    return

def test_bagle_mulens_set(plot_lc=False, plot_conv=False, verbose=False):
    """
    Test BAGLE --> Mulens as well as Mulens --> BAGLE 
    conversions for all combos of u0/piEE/piEN sign,
    as well as three sets of t0par (<, =, and > t0).
    """
    t_mjd = np.arange(57000 - 500, 57000 + 500, 1)

    kwargs = {'plot_lc' : plot_lc, 'plot_conv' : plot_lc, 'verbose' : verbose}

    print('set 1')
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57100, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57100, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57100, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57100, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57100, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57100, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57100, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57100, **kwargs)
    
    print('set 2')
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57100, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57100, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57100, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57100, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57100, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57100, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57100, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57100, **kwargs)
    
    print('set 3')
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57000, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57000, **kwargs)
    
    print('set 4')
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57000, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57000, **kwargs)
    
    print('set 5')
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 69900, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_bagle_to_mulens(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 69900, **kwargs)
    
    print('set 6')
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 69900, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_mulens_to_bagle(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 69900, **kwargs)


def test_bagle_helio_geo_set(plot_lc=False, plot_conv=False, verbose=False):
    """
    Test BAGLE's helio --> geo as well as geo --> helio
    conversions for all combos of u0/piEE/piEN sign,
    as well as three sets of t0par (<, =, and > t0).
    """
    t_mjd = np.arange(57000 - 500, 57000 + 500, 1)

    kwargs = {'plot_lc' : plot_lc, 'plot_conv' : plot_conv, 'verbose' : verbose}

#    one example to test (make figure for notes).
#    test_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57000, **kwargs)
#    pdb.set_trace()

    print('set 1') 

    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57100, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57100, **kwargs)
    #compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57100, **kwargs)
    #compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57100, **kwargs)
    #compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57100, **kwargs)
    #compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57100, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57100, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57100, **kwargs)
    
    print('set 2')
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57100, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57100, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57100, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57100, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57100, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57100, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57100, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57100, **kwargs)
    
    print('set 3')
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57000, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57000, **kwargs)
    
    print('set 4')
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 57000, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 57000, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 57000, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 57000, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 57000, **kwargs)
    
    print('set 5')
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 69900, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_bagle_helio_to_geo_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 69900, **kwargs)
    
    print('set 6')
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, 0.5, 300, -0.2, -0.1, 69900, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, 0.1, 69900, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, 0.1, 69900, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, 0.2, -0.1, 69900, **kwargs)
    compare_bagle_geo_to_helio_phot(259.0, -29.0, t_mjd, 57000, -0.5, 300, -0.2, -0.1, 69900, **kwargs)



################################################################################
###########  Stuff below is incomplete/wrong/not yet functional...  ############
################################################################################
# def compare_geo_geo():
#     t0_h, u0_h, tE_h, piEE_h, piEN_h = 57000, 0.5, 300, 0.2, 0.1
#
#     t0par1 = 57050
#     t0par2 = 56950
#
#     t0_g1, u0_g1, tE_g1, piEE_g1, piEN_g1 = fc.convert_helio_geo_phot(ra, dec, t0_h, u0_h,
#                                                                       tE_h, piEE_h, piEN_h,
#                                                                       t0par1, in_frame='helio',
#                                                                       murel_in='SL', murel_out='LS',
#                                                                       coord_in='EN', coord_out='tb')
#
#     t0_g2, u0_g2, tE_g2, piEE_g2, piEN_g2 = fc.convert_helio_geo_phot(ra, dec, t0_h, u0_h,
#                                                                       tE_h, piEE_h, piEN_h,
#                                                                       t0par2, in_frame='helio',
#                                                                       murel_in='SL', murel_out='LS',
#                                                                       coord_in='EN', coord_out='tb')
#
#     mag_mulens1 = get_phot_mulens(coords, t0_g1, u0_g1, tE_g1, piEE_g1, piEN_g1, t0par1, t_hjd)
#     mag_mulens2 = get_phot_mulens(coords, t0_g2, u0_g2, tE_g2, piEE_g2, piEN_g2, t0par2, t_hjd)
#     mag_bagle = get_phot_bagle(ra, dec, t0_h, u0_h, tE_h, piEE_h, piEN_h, t_mjd)
#
#     print(t0_g1, u0_g1, tE_g1, piEE_g1, piEN_g1, t0par1)
#     print(t0_g2, u0_g2, tE_g2, piEE_g2, piEN_g2, t0par2)
#
#     # For a formal test, require this to be smaller than some number.
#     # x = np.sum(np.abs(mag_mulens - mag_bagle))
#
#     fig, ax = plt.subplots(2, 1, sharex=True, num=3)
#     plt.clf()
#     fig, ax = plt.subplots(2, 1, sharex=True, num=3)
#     ax[0].plot(t_mjd, mag_mulens1, 'o')
#     ax[0].plot(t_mjd, mag_mulens2, 'o')
#     ax[0].plot(t_mjd, mag_bagle, '.')
#     ax[0].invert_yaxis()
#     ax[1].plot(t_mjd, mag_mulens1 - mag_bagle, '.')
#     ax[1].plot(t_mjd, mag_mulens2 - mag_bagle, '.')
#     ax[0].set_ylabel('Mag')
#     ax[1].set_ylabel('MM - Bagle')
#     ax[1].set_xlabel('MJD')
#     plt.show()
#     plt.pause(1)
# #    plt.close()
#
# def compare_geo():
#     t0_h, u0_h, tE_h, piEE_h, piEN_h = 57000, 0.5, 300, 0.2, 0.1
#
#     t0par_arr = np.arange(57000 - 720, 57000 + 720, 1)
#     t0_arr = np.zeros(len(t0par_arr))
#     u0_arr = np.zeros(len(t0par_arr))
#     tE_arr = np.zeros(len(t0par_arr))
#     piEE_arr = np.zeros(len(t0par_arr))
#     piEN_arr = np.zeros(len(t0par_arr))
#
#     mag_bagle = get_phot_bagle(ra, dec, t0_h, u0_h, tE_h, piEE_h, piEN_h, t_mjd)
#
#     fig, ax = plt.subplots(5, 1, sharex=True, num=3, figsize=(6,10))
#     plt.clf()
#     fig, ax = plt.subplots(5, 1, sharex=True, num=3, figsize=(6,10))
#     plt.subplots_adjust(left=0.25, top=0.98, bottom=0.1)
#
#     for ii, t0par in enumerate(t0par_arr):
#         t0_g, u0_g, tE_g, piEE_g, piEN_g = fc.convert_helio_geo_phot(ra, dec, t0_h, u0_h,
#                                                                           tE_h, piEE_h, piEN_h,
#                                                                           t0par, in_frame='helio',
#                                                                           murel_in='SL', murel_out='LS',
#                                                                           coord_in='EN', coord_out='tb')
#
#         t0_arr[ii] = t0_g
#         u0_arr[ii] = u0_g
#         tE_arr[ii] = tE_g
#         piEE_arr[ii] = piEE_g
#         piEN_arr[ii] = piEN_g
#
#         mag_mulens = get_phot_mulens(coords, t0_g, u0_g, tE_g, piEE_g, piEN_g, t0par, t_hjd)
#
# #        print(np.sum(np.abs(mag_mulens - mag_bagle)))
#
#     ax[0].plot(t0par_arr, t0_arr, 'k-', label='Geo projected')
#     ax[1].plot(t0par_arr, u0_arr, 'k-')
#     ax[2].plot(t0par_arr, tE_arr, 'k-')
#     ax[3].plot(t0par_arr, piEE_arr, 'k-')
#     ax[4].plot(t0par_arr, piEN_arr, 'k-')
#
#     ax[0].axhline(y=t0_h, ls=':', color='b', label='Helio')
#     ax[1].axhline(y=u0_h, ls=':', color='b')
#     ax[2].axhline(y=tE_h, ls=':', color='b')
#     ax[3].axhline(y=-1*piEE_h, ls=':', color='b')
#     ax[4].axhline(y=-1*piEN_h, ls=':', color='b')
#
#     ax[0].set_ylabel('$t_0$ (MJD)')
#     ax[1].set_ylabel('$u_0$')
#     ax[2].set_ylabel('$t_E$ (days)')
#     ax[3].set_ylabel('$\pi_{E,E}$ (LS)')
#     ax[4].set_ylabel('$\pi_{E,N}$ (LS)')
#     ax[4].set_xlabel('$t_{0,par}$ (MJD)')
#
#     ax[0].legend()
#     plt.show()



import math

import numpy as np
import joblib
import os
from astropy import units, units as u
from astropy.coordinates import SkyCoord, get_body_barycentric, get_body_barycentric_posvel, solar_system_ephemeris, \
    CartesianRepresentation
from astropy.time import Time

# Use the JPL ephemerides.
solar_system_ephemeris.set('jpl')

# Setup a parallax cache
try:
    cache_dir = os.environ['PARALLAX_CACHE_DIR']
except:
    cache_dir = os.path.dirname(__file__) + '/parallax_cache/'

# Setup cache with max size of 1 GB
if joblib.__version__ < '1.5':
    cache_memory = joblib.Memory(cache_dir, verbose=0, bytes_limit='1G')
    cache_memory.reduce_size()
else:
    cache_memory = joblib.Memory(cache_dir, verbose=0)
    cache_memory.reduce_size(bytes_limit='1G')

@cache_memory.cache()
def parallax_in_direction(RA, Dec, mjd, obsLocation='earth'):
    """
    | R.A. in degrees. (J2000)
    | Dec. in degrees. (J2000)
    | MJD

    Equations following MulensModel.
    """
    #print('parallax_in_direction: len(t) = ', len(mjd))

    # Munge inputs into astropy format.
    times = Time(mjd + 2400000.5, format='jd', scale='tdb')
    coord = SkyCoord(RA, Dec, unit=(units.deg, units.deg))

    direction = coord.cartesian.xyz.value
    north = np.array([0., 0., 1.])
    _east_projected = np.cross(north, direction) / np.linalg.norm(np.cross(north, direction))
    _north_projected = np.cross(direction, _east_projected) / np.linalg.norm(np.cross(direction, _east_projected))

    obs_pos = get_observer_barycentric(obsLocation, times)

    # Old Code for Heliocentric convention
    #sun_pos = get_body_barycentric(body='sun', time=times)
    #sun_obs_pos = sun_pos - obs_pos
    #pos = sun_obs_pos.xyz.T.to(units.au)

    # New Code for Barycentric convention
    bary_obs_pos = -obs_pos
    pos = bary_obs_pos.xyz.T.to(units.au)

    e = np.dot(pos, _east_projected)
    n = np.dot(pos, _north_projected)

    pvec = np.array([e.value, n.value]).T

    return pvec


def dparallax_dt_in_direction(RA, Dec, mjd, obsLocation='earth'):
    """
    R.A. in degrees. (J2000)
    Dec. in degrees. (J2000)
    MJD

    Equations following MulensModel.
    Time derivative --> units are yr^-1

    """
    # print('parallax_in_direction: len(t) = ', len(mjd))
    # Munge inputs into astropy format.
    times = Time(mjd + 2400000.5, format='jd', scale='tdb')
    coord = SkyCoord(RA, Dec, unit=(units.deg, units.deg))

    direction = coord.cartesian.xyz.value
    import pdb
    north = np.array([0., 0., 1.])
    _east_projected = np.cross(north, direction) / np.linalg.norm(np.cross(north, direction))
    _north_projected = np.cross(direction, _east_projected) / np.linalg.norm(np.cross(direction, _east_projected))

    obs_posvel = get_observer_barycentric(obsLocation, times, velocity=True)[1]
    sun_posvel = get_body_barycentric_posvel('Sun', times)[1]
    sun_obs_vel = sun_posvel - obs_posvel
    #pdb.set_trace()

    vel = sun_obs_vel.xyz.T.to(units.au / units.year)

    e = np.dot(vel, _east_projected)
    n = np.dot(vel, _north_projected)

    dpvec_dt = np.array([e.value, n.value]).T

    return dpvec_dt


def get_observer_barycentric(body, times, min_ephem_step=1, velocity=False):
    """
    Get the barycentric position of a satellite or other Solar System body
    using JPL emphemerides through the Horizon app.

    The ephemeris is queried at a decimated time step set by min_ephem_step
    (def=1 day) that must be 1 day or larger. The positions
    (and optionally velocities) are then interpolated onto the desired
    time array.

    Inputs
    ------
    body : str
        The name of the Solar System body. Must use the JPL Horizon
        naming scheme.

    times : astropy.time.Time array
        Array of times (astropy.time.core.Time) objects at which to
        fetch the position of the specified Solar System body. Times
        should be in the TDB scale.

    Optional Inputs
    ---------------
    min_ephem_step : int
        Minimum time step to query JPL in days. Must not be <1 and must
        be in integer days.

    veloctiy : bool
        If true, return both position and velocity vectors over time.

    Return
    ------
    coord : astropy.coordinates.CartesianRepresentation
        The xyz coordinates in the plane of the Solar System at the
        input times.
    """

    if body in solar_system_ephemeris.bodies:
        if velocity:
            obs_pos, obs_vel = get_body_barycentric_posvel(body=body, time=times)
        else:
            obs_pos = get_body_barycentric(body=body, time=times)
    else:
        # Figure out a cadence for the ephemerides, not smaller than 1 day.
        dt = np.median(np.diff(times)).jd
        if dt < min_ephem_step:
            dt = min_ephem_step

        # Get the date range, add some padding on each side.
        t_min = times.min()
        t_max = times.max()
        t_min.format = 'iso'
        t_max.format = 'iso'
        t_min = str(t_min - dt*u.day).split()[0]
        t_max = str(t_max + dt*u.day).split()[0]
        step = f'{dt:.0f}d'

        # Fetch the Horizons ephemeris.
        from astroquery.jplhorizons import Horizons
        obj = Horizons(id=body, location="@0", epochs={'start':t_min, 'stop':t_max, 'step':step})
        obj_data = obj.vectors()

        ephem_jd = obj_data['datetime_jd']

        # Interpolate to the actual time array.
        obj_x_at_t = np.interp(times.jd, ephem_jd, obj_data['x'].to('km')) * u.km
        obj_y_at_t = np.interp(times.jd, ephem_jd, obj_data['y'].to('km')) * u.km
        obj_z_at_t = np.interp(times.jd, ephem_jd, obj_data['z'].to('km')) * u.km

        if velocity:
            obj_vx_at_t = np.interp(times.jd, ephem_jd, obj_data['vx'].to('km/s')) * u.km / u.s
            obj_vy_at_t = np.interp(times.jd, ephem_jd, obj_data['vy'].to('km/s')) * u.km / u.s
            obj_vz_at_t = np.interp(times.jd, ephem_jd, obj_data['vz'].to('km/s')) * u.km / u.s

            obs_vel = CartesianRepresentation(obj_vx_at_t, obj_vy_at_t, obj_vz_at_t)

        obs_pos = CartesianRepresentation(obj_x_at_t, obj_y_at_t, obj_z_at_t)

    if velocity:
        return (obs_pos, obs_vel)
    else:
        return obs_pos


def sun_position(mjd, radians=False):
    """

    NAME:
          SUNPOS

    PURPOSE:
          To compute the RA and Dec of the Sun at a given date.

    INPUTS:
          mjd    - The modified Julian date of the day (and time), scalar or vector

    OUTPUTS:
          ra:
              | The right ascension of the sun at that date in DEGREES
              | double precision, same number of elements as jd
          dec:
              The declination of the sun at that date in DEGREES
          elong:
              Ecliptic longitude of the sun at that date in DEGREES.
          obliquity:
              the obliquity of the ecliptic, in DEGREES

    OPTIONAL INPUT KEYWORD:
          RADIAN [def=False] - If this keyword is set to True, then all output variables
          are given in Radians rather than Degrees

    NOTES:
          Patrick Wallace (Rutherford Appleton Laboratory, UK) has tested the
          accuracy of a C adaptation of the sunpos.pro code and found the
          following results.   From 1900-2100 SUNPOS  gave 7.3 arcsec maximum
          error, 2.6 arcsec RMS.  Over the shorter interval 1950-2050 the figures
          were 6.4 arcsec max, 2.2 arcsec RMS.

          The returned RA and Dec are in the given date's equinox.

          Procedure was extensively revised in May 1996, and the new calling
          sequence is incompatible with the old one.
    METHOD:
          Uses a truncated version of Newcomb's Sun.    Adapted from the IDL
          routine SUN_POS by CD Pike, which was adapted from a FORTRAN routine
          by B. Emerson (RGO).
    EXAMPLE:
          (1) Find the apparent RA and Dec of the Sun on May 1, 1982

          | IDL> jdcnv, 1982, 5, 1,0 ,jd      ;Find Julian date jd = 2445090.5
          | IDL> sunpos, jd, ra, dec
          | IDL> print,adstring(ra,dec,2)
          | 02 31 32.61  +14 54 34.9

          The Astronomical Almanac gives 02 31 32.58 +14 54 34.9 so the error
          in SUNPOS for this case is < 0.5".

          (2) Find the apparent RA and Dec of the Sun for every day in 1997

          | IDL> jdcnv, 1997,1,1,0, jd                ;Julian date on Jan 1, 1997
          | IDL> sunpos, jd+ dindgen(365), ra, dec    ;RA and Dec for each day

    MODIFICATION HISTORY:

          * Written by Michael R. Greason, STX, 28 October 1988.
          * Accept vector arguments, W. Landsman -     April,1989
          * Eliminated negative right ascensions - MRG, Hughes STX, 6 May 1992.
          * Rewritten using the 1993 Almanac.  Keywords added.  MRG, HSTX, 10 February 1994.
          * Major rewrite, improved accuracy, always return values in degrees - W. Landsman May, 1996
          * Added /RADIAN keyword; W. Landsman; August, 1997
          * Converted to IDL V5.0; W. Landsman; September 1997
          * Converted to python; J. R. Lu; August 2016
    """
    #  form time in Julian centuries from 1900.0
    t_obj = Time(mjd, format='mjd')
    t = (t_obj.jd - 2415020.0) / 36525.0

    #  form sun's mean longitude
    l = (279.696678 + ((36000.768925 * t) % 360.0)) * 3600.0

    #  allow for ellipticity of the orbit (equation of centre)
    #  using the Earth's mean anomaly ME
    me = 358.475844 + ((35999.049750 * t) % 360.0)
    ellcor = (6910.1 - 17.2 * t) * np.sin(np.radians(me)) + 72.3 * np.sin(
        np.radians(2.0 * me))
    l = l + ellcor

    # allow for the Venus perturbations using the mean anomaly of Venus MV
    mv = 212.603219 + ((58517.803875 * t) % 360.0)
    vencorr = 4.8 * np.cos(np.radians(299.1017 + mv - me)) + \
              5.5 * np.cos(np.radians(148.3133 + 2.0 * mv - 2.0 * me)) + \
              2.5 * np.cos(np.radians(315.9433 + 2.0 * mv - 3.0 * me)) + \
              1.6 * np.cos(np.radians(345.2533 + 3.0 * mv - 4.0 * me)) + \
              1.0 * np.cos(np.radians(318.1500 + 3.0 * mv - 5.0 * me))
    l += vencorr

    #  Allow for the Mars perturbations using the mean anomaly of Mars MM
    mm = 319.529425 + ((19139.858500 * t) % 360.0)
    marscorr = 2.0 * np.cos(np.radians(343.8883 - 2.0 * mm + 2.0 * me)) + \
               1.8 * np.cos(np.radians(200.4017 - 2.0 * mm + me))
    l += marscorr

    # Allow for the Jupiter perturbations using the mean anomaly of Jupiter MJ
    mj = 225.328328 + ((3034.6920239 * t) % 360.0)
    jupcorr = 7.2 * np.cos(np.radians(179.5317 - mj + me)) + \
              2.6 * np.cos(np.radians(263.2167 - mj)) + \
              2.7 * np.cos(np.radians(87.1450 - 2.0 * mj + 2.0 * me)) + \
              1.6 * np.cos(np.radians(109.4933 - 2.0 * mj + me))
    l += jupcorr

    # Allow for the Moons perturbations using the mean elongation of
    # the Moon from the Sun D
    d = 350.7376814 + ((445267.11422 * t) % 360.0)
    mooncorr = 6.5 * np.sin(np.radians(d))
    l += mooncorr

    # Allow for long period terms
    longterm = + 6.4 * np.sin(np.radians(231.19 + 20.20 * t))
    l += longterm
    l = (l + 2592000.0) % 1296000.0
    longmed = l / 3600.0

    # Allow for Aberration
    l -= 20.5

    # Allow for Nutation using the longitude of the Moons mean node OMEGA
    omega = 259.183275 - ((1934.142008 * t) % 360.0)
    l -= 17.2 * np.sin(np.radians(omega))

    # Form the True Obliquity
    oblt = 23.452294 - 0.0130125 * t + (
            9.2 * np.cos(np.radians(omega))) / 3600.0

    # Form Right Ascension and Declination
    l = l / 3600.0
    l_rad = np.radians(l)
    oblt_rad = np.radians(oblt)
    ra = np.arctan2(np.sin(l_rad) * np.cos(oblt_rad), np.cos(l_rad))

    if (len(ra) > 1):
        neg = np.where(ra < 0.0)[0]
        ra[neg] = ra[neg] + 2.0 * math.pi

    dec = np.arcsin(np.sin(l_rad) * np.sin(oblt_rad))

    if radians:
        oblt = oblt_rad
        longmed = np.radians(longmed)
    else:
        ra = np.degrees(ra)
        dec = np.degrees(dec)

    return ra, dec, longmed, oblt



import os, sys, math
import numpy as np
from astropy.table import Table, Column, MaskedColumn
import matplotlib.pyplot as plt
import astropy.constants as c
from astropy import units as u
from copy import deepcopy
from random import choice

days_per_year = 365.25

class Orbit(object):

    def oal2xy(self, t, mass=None, dist=None, accel=False):
        """
        This is for binaris orbiting each other and moving linearly together.
        Given a Orbit And a Linear motion, get the 2D position (x,y) at time epochs.
        Based on paper Koren et al. 2015
        
        Input:
        epochs -- epoch with detections

        Return:
        x -- x in arcsec
        y -- y in arcsec

        An example call is:

        orb = orbits.Orbit()
        orb.w = omega           # degree. The argument of periastron of the primary star's orbit. The secondary will be directly 180 degrees across
        orb.o = big           # degree (the longitude of the ascending node of the secondary star's orbit) - It is also the same as the primary
        orb.i = incl            # degree (Inclination angle of the system)
        
        orb.e = e_mag           # Eccentricity of the Keplerian Orbit
        orb.p = p               # Days (Orbital Period of the system)
        orb.tp = tp             # the time of the periastron passage in days
        orb.aleph = aleph       # semi-major axis of primary in arcsec
        orb.aleph2 = aleph      # semi-major axis of secondary in arcsec
        if accel=True:
            orb.ax = ax
            orb.ay = ay


            
        """

        
        if self.e <0 or self.e>1:
            return(-1e8, -1e8)
        meanMotion = 2.0 * math.pi / self.p

        ecc_sqrt = np.sqrt(1.0 - self.e**2)

        # Mean anomaly
        M = meanMotion * (t - self.tp)
        
        # Eccentric anomaly
        E = self.eccen_anomaly(M, self.e)

        # True Anomaly
        eta = 2* np.arctan(np.sqrt((1+self.e)/(1-self.e)) * np.tan(E/2))

        

        # elliptical rectangular coordinates Of the Orbit
        cos_E = np.cos(E)
        sin_E = np.sin(E)

        X = cos_E - self.e
        Y = ecc_sqrt * sin_E

        #----------
        # Calculate Thiele-Innes Constants for the Primary
        #----------
        cos_bigOm = np.cos(math.radians(self.o))
        sin_bigOm = np.sin(math.radians(self.o))
        cos_i = np.cos(math.radians(self.i))
        sin_i = np.sin(math.radians(self.i))
        cos_om = np.cos(math.radians(self.w))
        sin_om = np.sin(math.radians(self.w))
        

        self.conA = self.aleph * (cos_om * cos_bigOm  - sin_om * sin_bigOm * cos_i)
        self.conB = self.aleph * (cos_om * sin_bigOm  + sin_om * cos_bigOm * cos_i)
        self.conC = self.aleph * (sin_om * sin_i)
        self.conF = self.aleph * (-sin_om * cos_bigOm - cos_om * sin_bigOm * cos_i)
        self.conG = self.aleph * (-sin_om * sin_bigOm + cos_om * cos_bigOm * cos_i)
        self.conH = self.aleph * (cos_om * sin_i)


        
        #----------
        # Calculate Thiele-Innes Constants for the Secondary
        #----------
        
        cos_om2 = np.cos(math.radians(self.w+180))
        sin_om2 = np.sin(math.radians(self.w+180))

        
        self.conA2 = self.aleph2 * (cos_om2 * cos_bigOm  - sin_om2 * sin_bigOm * cos_i)
        self.conB2 = self.aleph2 * (cos_om2 * sin_bigOm  + sin_om2 * cos_bigOm * cos_i)
        self.conC2 = self.aleph2 * (sin_om2 * sin_i)
        self.conF2 = self.aleph2 * (-sin_om2 * cos_bigOm - cos_om2 * sin_bigOm * cos_i)
        self.conG2 = self.aleph2 * (-sin_om2 * sin_bigOm + cos_om2 * cos_bigOm * cos_i)
        self.conH2 = self.aleph2 * (cos_om2 * sin_i)



        # calculate x and y
        #The x and y that come out are the delta position of the star which you would then need to add to the center of mass 
        #x = (self.conB * X) + (self.conG * Y) + self.vx * (t-t0_com)/days_per_year + self.x0 
        #y = (self.conA * X) + (self.conF * Y) + self.vy * (t-t0_com)/days_per_year + self.y0
        #x2 = (self.conB2 * X) + (self.conG2 * Y) + self.vx * (t-t0_com)/days_per_year + (self.x0)
        #y2 = (self.conA2 * X) + (self.conF2 * Y) + self.vy * (t-t0_com)/days_per_year + (self.y0)
        #self.K1 = 3
        #self.gamma = 2
        #rv = self.K1 * (np.cos(math.radians(self.w)) + eta) + self.e * np.cos(math.radians(self.w)) + self.gamma
        #a1_num = self.K1  * self.p * np.sqrt(1-self.e**2)
        #a1_den = 2*np.pi * np.sin(self.i)
        #a1 = a1_num/a1_den
        #a1 = np.abs(a1)
        #rv is radial velocity
            
        
        # Unit conversions
        # r  from AU     to arcsec
        # v  from AU/yr  to mas/yr
        
        #x /= cc.dist
        #y /= cc.dist

        # x, y, x2, y2 = x/self.dS, y/self.dS, x2/self.dS, y2/self.dS
        x = (self.conB * X) + (self.conG * Y)
        y = (self.conA * X) + (self.conF * Y) 
        x2 = (self.conB2 * X) + (self.conG2 * Y) 
        y2 = (self.conA2 * X) + (self.conF2 * Y) 
        
        return (x, y, x2, y2)
        
    """
    Exerpted from orbits.py from gcworks
    """
    def kep2xyz(self, epochs):
        """
        Generates (r, v, a) in AU, AU/yr, and AU/yr^2 respectively from keplarian parameters. 
        
        epoch: numpy array
           Array of times (in years) that you would like to compute the parameters for
        
        mass: float or array-like
           Mass of primary object
        

        An example call is:

        orb = orbits.Orbit()
        orb.w = omega # [degrees]
        orb.o = bigOm # [degrees]
        orb.i = incl # [degrees]
        orb.e = e_mag # float between 0 and 1
        orb.p = p # [years]
        orb.t0 = t0 # [years] This is initial
        orb.mass = mass # [Msun]

        (r, v, a) = orb.kep2xyz(array([refTime]))
        
        """

        GM = self.mass * c.G.to("cm3/(Msun s2)")

        epoch_num = len(epochs)

        # meanMotion in radians per year
        meanMotion = 2.0 * math.pi / self.p
        
        # Semi-major axis in AU
        axis = (self.p**2 * self.mass)**(1.0/3.0)

        ecc_sqrt = np.sqrt(1.0 - self.e**2)        
        
        #switch angular quantities to radians
        w = math.radians(self.w)
        o = math.radians(self.o)
        i = math.radians(self.i)
        
        # Mean anomaly
        mean_anomaly = meanMotion * (epochs - self.t0)

        #----------
        # Now for each epoch we compute the x and y positions
        #----------

        # Eccentric anomaly
        E = self.eccen_anomaly(mean_anomaly, self.e)
            
        cos_E = np.cos(E)
        sin_E = np.sin(E)
        
        Edot = meanMotion / (1.0 - (self.e * cos_E))

        X = cos_E - self.e
        Y = ecc_sqrt * sin_E

        #----------
        # Calculate Thiele-Innes Constants
        #----------
        cos_bigOm = np.cos(o)
        sin_bigOm = np.sin(o)
        cos_i = np.cos(i)
        sin_i = np.sin(i)
        
        cos_om = np.cos(w)
        sin_om = np.sin(w)

        self.conA = axis * (cos_om * cos_bigOm  - sin_om * sin_bigOm * cos_i)
        self.conB = axis * (cos_om * sin_bigOm  + sin_om * cos_bigOm * cos_i)
        self.conC = axis * (sin_om * sin_i)
        self.conF = axis * (-sin_om * cos_bigOm - cos_om * sin_bigOm * cos_i)
        self.conG = axis * (-sin_om * sin_bigOm + cos_om * cos_bigOm * cos_i)
        self.conH = axis * (cos_om * sin_i)
        
        # initialize zero arrays for r, v, and a
        r = np.zeros((epoch_num, 3), dtype='float64')
        v = np.zeros((epoch_num, 3), dtype='float64')
        a = np.zeros((epoch_num, 3), dtype='float64')

        r[:,0] = (self.conB * X) + (self.conG * Y)
        r[:,1] = (self.conA * X) + (self.conF * Y)
        r[:,2] = (self.conC * X) + (self.conH * Y)
        
        v[:,0] = Edot * ((-self.conB * sin_E) + (self.conG * ecc_sqrt * cos_E))
        v[:,1] = Edot * ((-self.conA * sin_E) + (self.conF * ecc_sqrt * cos_E))
        v[:,2] = Edot * ((-self.conC * sin_E) + (self.conH * ecc_sqrt * cos_E))
        
    
        # Calculate accleration
        for ii in range(epoch_num):
            rmag_cm = (np.sqrt( (r[ii,:]**2).sum() )*(u.au)).to("cm").value
            a[ii,:] = -GM * (r[ii,:]*(u.au)).to("cm").value / rmag_cm**3
        
        # from cm/s^2 to AU/yr^2
        a = (a*(u.cm/u.s**2)).to("au/yr2").value

        return (r, v, a)


    def eccen_anomaly(self, m, ecc, thresh=1e-10):
        """
        Calculates the eccentricity anomaly
        
        m: numpy array
            Mean anomalies
            
        ecc: float between 0-1
            The eccentricity of the orbit
        """
        # set default values

        if (ecc < 0. or ecc >= 1.):
            raise Exception('Eccentricity must be >= 0 and < 1')
            
        #
        # Range reduction of m to -pi < m <= pi
        #
        mx = m.copy()

        ## ... m > pi
        zz = (np.where(mx > math.pi))[0]
        mx[zz] = mx[zz] % (2.0 * math.pi)
        zz = (np.where(mx > math.pi))[0]
        mx[zz] = mx[zz] - (2.0 * math.pi)
        
        # ... m < -pi
        zz = (np.where(mx <= -math.pi))[0]
        mx[zz] = mx[zz] % (2.0 * math.pi)
        zz = (np.where(mx <= -math.pi))[0]
        mx[zz] = mx[zz] + (2.0 * math.pi)

        #
        # Bail out for circular orbits...
        #
        if (ecc == 0.0):
            return mx

        aux   = (4.0 * ecc) + 0.50
        alpha = (1.0 - ecc) / aux

        beta = mx/(2.0*aux)
        aux = np.sqrt(beta**2 + alpha**3)
   
        z=beta+aux
        zz=(np.where(z <= 0.0))[0]
        z[zz]=beta[zz]-aux[zz]

        test=abs(z)**0.3333333333333333

        z =  test.copy()
        zz = (np.where(z < 0.0))[0]
        z[zz] = -z[zz]

        s0=z-alpha/z
        s1 = s0-(0.0780 * s0**5) / (1.0 + ecc)
        e0 = mx + ecc*((3.0 * s1) - (4.0 * s1**3))

        se0=np.sin(e0)
        ce0=np.cos(e0)

        f  = e0 - (ecc*se0) - mx
        f1 = 1.0 - (ecc*ce0)
        f2 = ecc*se0
        f3 = ecc*ce0
        f4 = -1.0 * f2
        u1 = -1.0 * f/f1
        u2 = -1.0 * f/(f1 + 0.50*f2*u1)
        u3 = -1.0 * f/(f1 + 0.50*f2*u2
                 + 0.166666666666670*f3*u2*u2)
        u4 = -1.0 * f/(f1 + 0.50*f2*u3
                 + 0.166666666666670*f3*u3*u3
                 + 0.0416666666666670*f4*u3**3)

        eccanom=e0+u4

        zz = (np.where(eccanom >= 2.00*math.pi))[0]
        eccanom[zz]=eccanom[zz]-2.00*math.pi
        zz = (np.where(eccanom < 0.0))[0]
        eccanom[zz]=eccanom[zz]+2.00*math.pi

        # Now get more precise solution using Newton Raphson method
        # for those times when the Kepler equation is not yet solved
        # to better than 1e-10
        # (modification J. Wilms)

        mmm = mx.copy()
        ndx = (np.where(mmm < 0.))[0]
        mmm[ndx] += (2.0 * math.pi)
        diff = eccanom - ecc*np.sin(eccanom) - mmm

        ndx = (np.where(abs(diff) > 1e-10))[0]
        for i in ndx:
            # E-e sinE-M
            fe = eccanom[i]-ecc*np.sin(eccanom[i])-mmm[i]
            # f' = 1-e*cosE
            fs = 1.0 - ecc*np.cos(eccanom[i])
            oldval=eccanom[i]
            eccanom[i]=oldval-fe/fs

            loopCount = 0
            while (abs(oldval-eccanom[i]) > thresh):
                # E-e sinE-M
                fe = eccanom[i]-ecc*np.sin(eccanom[i])-mmm[i]
                # f' = 1-e*cosE
                fs = 1.0 - ecc*np.cos(eccanom[i])
                oldval=eccanom[i]
                eccanom[i]=oldval-fe/fs
                loopCount += 1
                
                if (loopCount > 10**6):
                    msg = 'eccen_anomaly: Could not converge for e = %f' % ecc
                    raise EccAnomalyError(msg)

            while (eccanom[i] >=  math.pi):
                eccanom[i] = eccanom[i] - (2.0 * math.pi)
                
            while (eccanom[i] < -math.pi ):
                eccanom[i] = eccanom[i] + (2.0 * math.pi)

        return eccanom
    
class EccAnomalyError(Exception):
    def __init__(self, message):
        self.message = message

def a_to_P(mass, a):
    """
    Goes from semimajor axis in AU to period in years
    
    Parameters
    ---------- 
    mass: float or array-like
        Primary object mass in Msun.
    
    a: float or array-like
        Semimajor axis in AU.
        
    Returns
    ---------- 
    period: float or array-like
        Orbital period in years.
    """
    
    G_units = c.G.to("AU3/(M_sun*year2)").value
    period = (a**3*4*(np.pi**2)/G_units/mass)**(1/2)
    return period

def add_positions(ss):
    """
    Adds x and y positions randomly in a box of length and width 40000 AU for each system.
    
    Parameters
    ----------   
    ss: astropy table
        Star system table without positions
        
    Returns
    -------
    ss_temp: astropy table
        Star system table with positions added
    """
    ss_temp = deepcopy(ss)
    
    ss_temp.add_column( Column(np.zeros(len(ss_temp), dtype=float), name='x', description='AU') )
    ss_temp.add_column( Column(np.zeros(len(ss_temp), dtype=float), name='y', description='AU') )
    
    sign_x = np.array([choice([-1,1]) for i in range(len(ss_temp))])
    sign_y = np.array([choice([-1,1]) for i in range(len(ss_temp))])
    ss_temp['x'] = sign_x*20000*np.random.rand(len(ss_temp))
    ss_temp['y'] = sign_y*20000*np.random.rand(len(ss_temp))
    
    return ss_temp

def add_mult_positions(companions, ss_pos, logAge, add_velocities = False):
    """
    Adds x and y positions of multiple companions by transforming keplerian parameters to xyz in AU
    using code origially from gcworks and random initial times. Then adding them to the random posiiton of the primary object.
    
    Parameters
    ----------    
    ss_pos: astropy table
        Star system table with positions added with add_positions()
    
    companion: astropy table
        Companion table without positions
    
    logAge: float or int
        Log of age of cluster with age in years
        
    Returns
    -------
    companion_temp: astropy table
        Companion table with positions added
        
    """
    companions_temp = deepcopy(companions)
    
    companions_temp.add_column( Column(np.zeros(len(companions_temp), dtype=float), name='x', description='AU') )
    companions_temp.add_column( Column(np.zeros(len(companions_temp), dtype=float), name='y', description='AU') )
    
    if add_velocities == True:
        companions_temp.add_column( Column(np.zeros(len(companions_temp), dtype=float), name='v_x', description='AU/yr') )
        companions_temp.add_column( Column(np.zeros(len(companions_temp), dtype=float), name='v_y', description='AU/yr') )
        
    orb = Orbit()
    for i in companions_temp:
        orb.w = i['omega'] #degrees
        orb.o = i['Omega'] #degrees
        orb.i = i['i'] #degrees
        orb.e = i['e'] #between 0 and 1
        orb.p = a_to_P(ss_pos[i['system_idx']]['systemMass'],10**i['log_a']) #year
        orb.t0 = (10**logAge)*np.random.rand() #year
        orb.mass = ss_pos[i['system_idx']]['mass']
        
        (r, v, a) = orb.kep2xyz(np.array([10**logAge]))
               
        x = r[0][0]
        y = r[0][1]
        
        #putting positions relative to primary object
        i['x'] = ss_pos[i['system_idx']]['x'] + x #AU
        i['y'] = ss_pos[i['system_idx']]['y'] + y #AU
        
        if add_velocities == True:
            i['v_x'] = v[0][0] #AU/yr
            i['v_y'] = v[0][1] #AU/yr
    
    
    return companions_temp

def distance_to_center_of_mass(ss_pos, companions_pos):
    """
    Adds extra column to star system and companions table with x and y distance to the center of mass in AU.
    Assumes hierarchical triples (two closest stars orbit their center of mass and triple orbits them) and no quads+.
    
    Parameters
    ----------    
    ss_pos: astropy table
        Star system table with positions added with add_positions()
    
    companion_pos: astropy table
        Companion table with positions added with add_mult_positions()
        
    Results
    -------- 
    ss_pos_temp: astropy table
        Star system table with distance to center of mass in AU added
    
    companion_pos_temp: astropy table
        Companion table with distance to center of mass in AU added
    """
    
    ss_pos_temp = deepcopy(ss_pos)
    companions_pos_temp = deepcopy(companions_pos)
    
    ss_pos_temp.add_column( Column(np.zeros(len(ss_pos_temp), dtype=float), name='com_x') )
    ss_pos_temp.add_column( Column(np.zeros(len(ss_pos_temp), dtype=float), name='com_y') )
    companions_pos_temp.add_column( Column(np.zeros(len(companions_pos_temp), dtype=float), name='com_x') )
    companions_pos_temp.add_column( Column(np.zeros(len(companions_pos_temp), dtype=float), name='com_y') )
    
    companions_pos_temp.sort(['system_idx','log_a'])
    
    for i in range(len(ss_pos_temp)):
        if ss_pos_temp[i]['isMultiple'] == True:
            companion_indicies = np.where(companions_pos_temp['system_idx'] == i)[0]
            
            primary_mass = ss_pos_temp[i]['mass']
            companion_masses = companions_pos_temp[companion_indicies]['mass']
            
            primary_x = ss_pos_temp[i]['x']
            companion_x = companions_pos_temp[companion_indicies]['x']
            com_x = (primary_mass*primary_x + companion_masses[0]*companion_x[0])/(primary_mass + companion_masses[0])
            
            primary_y = ss_pos_temp[i]['y']
            companion_y = companions_pos_temp[companion_indicies]['y']
            com_y = (primary_mass*primary_y + companion_masses[0]*companion_y[0])/(primary_mass + companion_masses[0])
            
            ss_pos_temp[i]['com_x'] = com_x - primary_x
            companions_pos_temp[companion_indicies[0]]['com_x'] = com_x - companion_x[0]
            ss_pos_temp[i]['com_y'] = com_y - primary_y
            companions_pos_temp[companion_indicies[0]]['com_y'] = com_y - companion_y[0]
            
            # Assumes hierarchical triples 
            if len(companion_indicies) == 2:
                center_mass = primary_mass + companion_masses[0]
                com_x_out = (center_mass*com_x + companion_masses[1]*companion_x[1])/(center_mass + companion_masses[1])
                com_y_out = (center_mass*com_y + companion_masses[1]*companion_y[1])/(center_mass + companion_masses[1])
                
                companions_pos_temp[companion_indicies[1]]['com_x'] = com_x_out - companion_x[1]
                companions_pos_temp[companion_indicies[1]]['com_y'] = com_y_out - companion_y[1]
                
    return ss_pos_temp, companions_pos_temp
                
            
def plot_projected_cluster(ss_pos, companions_pos):
    """
    Plots projected cluster with lines between companions and primary stars
    
    Parameters
    ----------   
    ss_pos: astropy table
        Star system table with positions added with add_positions()
    
    companion_pos: astropy table
        Companion table with positions added with add_mult_positions()
    """
    plt.figure(figsize=(10,10))
    plt.plot(ss_pos['x'], ss_pos['y'],linestyle='none',marker='o' )
    plt.plot(companions_pos['x'], companions_pos['y'],linestyle='none',marker='.' )
    
    #makes lines between companion and primary star
    for i in companions_pos:
        plt.plot([i['x'], ss_pos[i['system_idx']]['x']],[i['y'], ss_pos[i['system_idx']]['y']],color='grey',linewidth=1)
        
    plt.xlabel("x (AU)")
    plt.ylabel("y (AU)")
    plt.show()
    
    return

def plot_companion_orbit(ss, companions_pos, logAge, t0 = None, system = None):
    """
    Plots the orbit of one system assuming the primary object is at (0,0). By default random companion and initial time.
    
    Parameters
    ----------       
    ss: astropy table
        Star system table (does not matter if it has positions or not)
    
    companion_pos: astropy table
        Companion table with positions added with add_mult_positions()
    
    logAge: float or int
        Log of age of cluster with age in years
        
    t0: float or int, optional
        Initial time of creation of the system in years.
        Default random.
    
    system: int, optional
        Index of desired companion in companion_pos table.
        Default random.
    """
    
    if system == None:
        system = np.random.randint(len(companions_pos))
    if t0 == None:
        t0 = (10**logAge)*np.random.rand() #year

    companion = companions_pos[system]
    
    orb = Orbit()
    orb.w = companion['omega'] #degrees
    orb.o = companion['Omega'] #degrees
    orb.i = companion['i'] #degrees
    orb.e = companion['e'] #between 0 and 1
    orb.p = a_to_P(ss[companion['system_idx']]['mass'],10**companion['log_a']) #year
    orb.t0 = t0 #year
        
    (r, v, a) = orb.kep2xyz(np.linspace(1, 10**logAge), mass=ss[companion['system_idx']]['mass'])
        
    for i in r:
        plt.plot(i[0],i[1], marker='.', color = 'blue')
        if i[0] == r[-1][0]:
            plt.plot(i[0],i[1], marker='*', color = 'gold', markersize=15)
            
    plt.xlabel("x (AU)")
    plt.ylabel("y (AU)")
    plt.show()
    
    return


import numpy as np
import pylab as plt
from bagle import model
from bagle import sensitivity

def test_fisher_matrix(plot=False):
    """Test PSPL_Phot_Par_Param1 and that
    adding data reduces errors. 
    """
    params = {}
    params['t0'] = 60000.0 # MJD
    params['u0_amp'] = 0.1 # thetaE
    params['tE'] = 150.0   # days
    params['piEE'] = 0.04
    params['piEN'] = 0.03
    params['b_sff'] = 0.1
    params['mag_src'] = 19.0
    raL = 17.30 * 15.  # Bulge R.A.
    decL = -29.0

    # Make some fake data. Only sample up until tE.

    pspl_par_in = model.PSPL_Phot_Par_Param1(params['t0'],
                                             params['u0_amp'],
                                             params['tE'],
                                             params['piEE'],
                                             params['piEN'],
                                             params['b_sff'],
                                             params['mag_src'],
                                             raL=raL, decL=decL)

    # Simulate
    # photometric observations every 1 day and
    # for the bulge observing window. Observations missed
    # for 125 days out of 365 days for photometry and missed
    # for 245 days out of 365 days for astrometry.
    t_phot = np.array([], dtype=float)
    
    for year_start in np.arange(params['t0'] - 1000, params['t0'] + 1000, 365.25):
        phot_win = 240.0
        phot_start = (365.25 - phot_win) / 2.0
        t_phot_new = np.arange(year_start + phot_start,
                               year_start + phot_start + phot_win, 1)
        t_phot = np.concatenate([t_phot, t_phot_new])

    # Make the photometric observations.
    # Assume 0.05 mag photoemtric errors at I=19.
    # This means Signal = 400 e- at I=19.
    flux0 = 4000.0
    imag0 = 19.0
    imag_obs = pspl_par_in.get_photometry(t_phot)
    flux_obs = flux0 * 10 ** ((imag_obs - imag0) / -2.5)
    flux_obs_err = flux_obs ** 0.5
    flux_obs += np.random.randn(len(t_phot)) * flux_obs_err
    imag_obs = -2.5 * np.log10(flux_obs / flux0) + imag0
    imag_obs_err = 1.087 / flux_obs_err

    # Plot the data.
    if plot:
        plt.close(1)
        plt.figure(1)
        plt.clf()
        plt.errorbar(t_phot, imag_obs, yerr=imag_obs_err, fmt='k.')
        plt.gca().invert_yaxis()

    params_fixed = {'raL': raL, 'decL': decL}

    # Check that the errors go down as we add data after the peak.
    # Data only up to tE:
    idx1 = np.where(t_phot < params['t0'])[0]
    cov_mat1 = sensitivity.fisher_matrix(t_phot[idx1], imag_obs_err[idx1],
                                         model.PSPL_Phot_Par_Param1,
                                         params, params_fixed)
    err_on_param1 = np.sqrt(np.diagonal(cov_mat1))
 
    # Data all the way through.
    cov_mat2 = sensitivity.fisher_matrix(t_phot, imag_obs_err,
                                         model.PSPL_Phot_Par_Param1,
                                         params, params_fixed)
    err_on_param2 = np.sqrt(np.diagonal(cov_mat2))


    param_names = list(params.keys())

    for pp in range(len(param_names)):
        assert err_on_param2[pp] < err_on_param1[pp]
        # print(f'Error1 on {param_names[pp]:10s} = {err_on_param1[pp]:.3f} at value {params[param_names[pp]]:.3f}')
        
    return
    

def test_fisher_matrix2(plot=False):
    """Test PSPL_Phot_Par_Param3 and that
    adding data reduces errors. 
    """

    model_class = model.PSPL_Phot_Par_Param3
    params = {}
    params['t0'] = 60000.0 # MJD
    params['u0_amp'] = 0.1 # thetaE
    params['log_tE'] = np.log10(150.0)   # log(days)
    params['log_piE'] = np.log10(0.04)   
    params['phi_muRel'] = 30.0           # degrees
    params['b_sff'] = 0.1
    params['mag_base'] = 16.5
    
    params_fixed = {}
    params_fixed['raL'] = 17.30 * 15.  # Bulge R.A.
    params_fixed['decL'] = -29.0

    # Make some fake data. Only sample up until tE.
    pspl_par_in = model.get_model(model_class, params, params_fixed)

    # Simulate
    # photometric observations every 1 day and
    # for the bulge observing window. Observations missed
    # for 125 days out of 365 days for photometry and missed
    # for 245 days out of 365 days for astrometry.
    t_phot = np.array([], dtype=float)
    
    for year_start in np.arange(params['t0'] - 1000, params['t0'] + 1000, 365.25):
        phot_win = 240.0
        phot_start = (365.25 - phot_win) / 2.0
        t_phot_new = np.arange(year_start + phot_start,
                               year_start + phot_start + phot_win, 1)
        t_phot = np.concatenate([t_phot, t_phot_new])

    # Make the photometric observations.
    # Assume 0.05 mag photoemtric errors at I=19.
    # This means Signal = 400 e- at I=19.
    flux0 = 4000.0
    imag0 = 19.0
    imag_obs = pspl_par_in.get_photometry(t_phot)
    flux_obs = flux0 * 10 ** ((imag_obs - imag0) / -2.5)
    flux_obs_err = flux_obs ** 0.5
    flux_obs += np.random.randn(len(t_phot)) * flux_obs_err
    imag_obs = -2.5 * np.log10(flux_obs / flux0) + imag0
    imag_obs_err = 1.087 / flux_obs_err

    # Plot the data.
    if plot:
        plt.close(2)
        plt.figure(2)
        plt.clf()
        plt.errorbar(t_phot, imag_obs, yerr=imag_obs_err, fmt='r.')
        plt.gca().invert_yaxis()


    # Check that the errors go down as we add data after the peak.
    # Data only up to tE:
    idx1 = np.where(t_phot < params['t0'])[0]
    cov_mat1 = sensitivity.fisher_matrix(t_phot[idx1], imag_obs_err[idx1],
                                         model_class,
                                         params, params_fixed)
    err_on_param1 = np.sqrt(np.diagonal(cov_mat1))
 
    # Data all the way through.
    cov_mat2 = sensitivity.fisher_matrix(t_phot, imag_obs_err,
                                         model_class,
                                         params, params_fixed)
    err_on_param2 = np.sqrt(np.diagonal(cov_mat2))


    param_names = list(params.keys())

    for pp in range(len(param_names)):
        assert err_on_param2[pp] < err_on_param1[pp]
        
    return
    
def test_compare_err_phot_param2_vs_param3():
    """
    Compare errors from models with parameters vs. log parameters. 
    """
    model_class2 = model.PSPL_Phot_Par_Param2
    params2 = {}
    params2['t0'] = 60000.0 # MJD
    params2['u0_amp'] = 0.1 # thetaE
    params2['tE'] = 150.0   # days)
    params2['piE_E'] = -0.05
    params2['piE_N'] = 0.11
    params2['b_sff'] = 0.1
    params2['mag_base'] = 16.5
    
    model_class3 = model.PSPL_Phot_Par_Param3
    params3 = {}
    params3['t0'] = 60000.0 # MJD
    params3['u0_amp'] = 0.1 # thetaE
    params3['log_tE'] = np.log10(150.0)   # log(days)
    params3['log_piE'] = np.log10(np.linalg.norm([params2['piE_E'],
                                                  params2['piE_N']]))
    params3['phi_muRel'] = np.rad2deg(np.arctan2(params2['piE_E'],
                                                 params2['piE_N']))
    params3['b_sff'] = 0.1
    params3['mag_base'] = 16.5

    
    params_fixed = {}
    params_fixed['raL'] = (17.0 + (49.0 / 60.) + (51.38 / 3600.0)) * 15.0  # degrees
    params_fixed['decL'] = -35 + (22.0 / 60.0) + (28.0 / 3600.0)
    

    t_mod = np.arange(params2['t0'] - 1000, params2['t0'] + 1000, 2)

    ##########
    # Compare two models with same properties
    # recast in different parameters.
    ##########
    mod2 = model.get_model(model_class2, params2, params_fixed)
    mod3 = model.get_model(model_class3, params3, params_fixed)

    # Make the photometric observations.
    # Assume 0.05 mag photoemtric errors at I=19.
    # This means Signal = 400 e- at I=19.
    flux0 = 4000.0
    imag0 = 19.0
    imag_obs = mod2.get_photometry(t_mod)
    flux_obs = flux0 * 10 ** ((imag_obs - imag0) / -2.5)
    flux_obs_err = flux_obs ** 0.5
    imag_obs_err = 1.087 / flux_obs_err

    # Assume same errors for both models. 
    fish2 = sensitivity.fisher_matrix(t_mod, imag_obs_err,
                                      model_class2, params2, params_fixed)
    fish3 = sensitivity.fisher_matrix(t_mod, imag_obs_err,
                                      model_class3, params3, params_fixed)
    
    err_on_param2 = np.sqrt(np.diagonal(fish2))
    err_on_param3 = np.sqrt(np.diagonal(fish3))

    for pp, key in enumerate(params2):
        print(f'Error2 on {key:10s} = {err_on_param2[pp]:.3f}')

        if key == 'tE':
            print(f'Error2 on {"log_tE":10s} = {np.log10(err_on_param2[pp]):.3f}')

        if key == 'piE_N':
            err_piE = np.hypot(err_on_param2[pp], err_on_param2[pp-1])
            print(f'Error2 on {"piE":10s} = {err_piE:.3f}')
            print(f'Error2 on {"log_piE":10s} = {np.log10(err_piE):.3f}')
            

    print('---')
    for pp, key in enumerate(params3):
        print(f'Error3 on {key:10s} = {err_on_param3[pp]:.3f}')

        if key == 'log_tE':
            print(f'Error3 on {"tE":10s} = {10**err_on_param3[pp]:.3f}')

        if key == 'log_piE':
            err_piE = err_on_param3[pp]
            print(f'Error3 on {"piE":10s} = {10**err_on_param3[pp]:.3f}')
            
            

    return

    
